/* redefac.h generated from autoconf.h.in on Thu May 20 12:36:02 MEST 1999  */

#ifndef HAVE_DOPRNT
#define HAVE_DOPRNT 0
#endif


#ifndef HAVE_ST_BLKSIZE
#define HAVE_ST_BLKSIZE 0
#endif


#ifndef HAVE_STRFTIME
#define HAVE_STRFTIME 0
#endif


#ifndef HAVE_SYS_WAIT_H
#define HAVE_SYS_WAIT_H 0
#endif


#ifndef HAVE_TM_ZONE
#define HAVE_TM_ZONE 0
#endif


#ifndef HAVE_TZNAME
#define HAVE_TZNAME 0
#endif


#ifndef HAVE_UTIME_NULL
#define HAVE_UTIME_NULL 0
#endif


#ifndef HAVE_VPRINTF
#define HAVE_VPRINTF 0
#endif


#ifndef HAVE_OPENLOG_LOG_PERROR
#define HAVE_OPENLOG_LOG_PERROR 0
#endif


#ifndef HAVE_CMSGHDR
#define HAVE_CMSGHDR 0
#endif


#ifndef HAVE_OPTHDR
#define HAVE_OPTHDR 0
#endif


#ifndef HAVE_MSG_WAITALL
#define HAVE_MSG_WAITALL 0
#endif


#ifndef NEED_SYS_SOCKIO_H
#define NEED_SYS_SOCKIO_H 0
#endif


#ifndef HAVE_MALLOC_OPTIONS
#define HAVE_MALLOC_OPTIONS 0
#endif


#ifndef HAVE_PROGNAME
#define HAVE_PROGNAME 0
#endif


#ifndef NEED_GETSOCKOPT_CAST
#define NEED_GETSOCKOPT_CAST 0
#endif


#ifndef HAVE_FAULTY_BINDPROTO
#define HAVE_FAULTY_BINDPROTO 0
#endif


#ifndef HAVE_FAULTY_CONNECTPROTO
#define HAVE_FAULTY_CONNECTPROTO 0
#endif


#ifndef HAVE_FAULTY_READVPROTO
#define HAVE_FAULTY_READVPROTO 0
#endif


#ifndef HAVE_LINUX_SOCKADDR_TYPE
#define HAVE_LINUX_SOCKADDR_TYPE 0
#endif


#ifndef HAVE_RECVFROM_CHAR
#define HAVE_RECVFROM_CHAR 0
#endif


#ifndef HAVE_SENDTO_ALT
#define HAVE_SENDTO_ALT 0
#endif


#ifndef NEED_SA_RESTART
#define NEED_SA_RESTART 0
#endif


#ifndef NEED_AF_LOCAL
#define NEED_AF_LOCAL 0
#endif


#ifndef NEED_EXIT_FAILURE
#define NEED_EXIT_FAILURE 0
#endif


#ifndef HAVE_INET_ATON
#define HAVE_INET_ATON 0
#endif


#ifndef NEED_ACCEPTLOCK
#define NEED_ACCEPTLOCK 0
#endif


#ifndef HAVE_DLFCN_H
#define HAVE_DLFCN_H 0
#endif


#ifndef HAVE_SYS_SOCKIO_H
#define HAVE_SYS_SOCKIO_H 0
#endif


#ifndef NEED_DYNA_RTLD
#define NEED_DYNA_RTLD 0
#endif


#ifndef HAVE_SO_SNDLOWAT
#define HAVE_SO_SNDLOWAT 0
#endif


#ifndef HAVE_SIGNAL_SIGINFO
#define HAVE_SIGNAL_SIGINFO 0
#endif


#ifndef HAVE_GETHOSTBYNAME2
#define HAVE_GETHOSTBYNAME2 0
#endif


#ifndef HAVE_SENDMSG_DEADLOCK
#define HAVE_SENDMSG_DEADLOCK 0
#endif


#ifndef HAVE_NOMALLOC_REALLOC
#define HAVE_NOMALLOC_REALLOC 0
#endif


#ifndef HAVE_DEFECT_RECVMSG
#define HAVE_DEFECT_RECVMSG 0
#endif


#ifndef HAVE_SOLARIS_2_5_1
#define HAVE_SOLARIS_2_5_1 0
#endif


#ifndef HAVE_SOLARIS_BUGS
#define HAVE_SOLARIS_BUGS 0
#endif


#ifndef HAVE_LINUX_BUGS
#define HAVE_LINUX_BUGS 0
#endif


#ifndef HAVE_BROKEN_VSPRINTF
#define HAVE_BROKEN_VSPRINTF 0
#endif


#ifndef HAVE_HOST_TYPE
#define HAVE_HOST_TYPE 0
#endif


#ifndef NEED_UCBINCLUDE_SYS_IOCTL
#define NEED_UCBINCLUDE_SYS_IOCTL 0
#endif


#ifndef HAVE_PSOCKLEN_T
#define HAVE_PSOCKLEN_T 0
#endif


#ifndef HAVE_VOID_SOCKADDR
#define HAVE_VOID_SOCKADDR 0
#endif


#ifndef HAVE_RECV_LEN_INT
#define HAVE_RECV_LEN_INT 0
#endif


#ifndef HAVE_STRUCT_IPOPTS
#define HAVE_STRUCT_IPOPTS 0
#endif


#ifndef HAVE_BROKEN_INET_NTOA
#define HAVE_BROKEN_INET_NTOA 0
#endif


#ifndef HAVE_DEC_PROTO
#define HAVE_DEC_PROTO 0
#endif


#ifndef HAVE_EXTRA_OSF_SYMBOLS
#define HAVE_EXTRA_OSF_SYMBOLS 0
#endif


#ifndef HAVE_SIG_ATOMIC_T
#define HAVE_SIG_ATOMIC_T 0
#endif


#ifndef HAVE_NO_RESOLVESTUFF
#define HAVE_NO_RESOLVESTUFF 0
#endif


#ifndef HAVE_NONULL_FREE
#define HAVE_NONULL_FREE 0
#endif


#ifndef HAVE_BZERO
#define HAVE_BZERO 0
#endif


#ifndef HAVE_DAEMON
#define HAVE_DAEMON 0
#endif


#ifndef HAVE_DIFFTIME
#define HAVE_DIFFTIME 0
#endif


#ifndef HAVE_GETDTABLESIZE
#define HAVE_GETDTABLESIZE 0
#endif


#ifndef HAVE_HSTRERROR
#define HAVE_HSTRERROR 0
#endif


#ifndef HAVE_INET_PTON
#define HAVE_INET_PTON 0
#endif


#ifndef HAVE_ISSETUGID
#define HAVE_ISSETUGID 0
#endif


#ifndef HAVE_MEMMOVE
#define HAVE_MEMMOVE 0
#endif


#ifndef HAVE_SELECT
#define HAVE_SELECT 0
#endif


#ifndef HAVE_SETPROCTITLE
#define HAVE_SETPROCTITLE 0
#endif


#ifndef HAVE_SNPRINTF
#define HAVE_SNPRINTF 0
#endif


#ifndef HAVE_SOCKATMARK
#define HAVE_SOCKATMARK 0
#endif


#ifndef HAVE_SOCKET
#define HAVE_SOCKET 0
#endif


#ifndef HAVE_STRCSPN
#define HAVE_STRCSPN 0
#endif


#ifndef HAVE_STRDUP
#define HAVE_STRDUP 0
#endif


#ifndef HAVE_STRERROR
#define HAVE_STRERROR 0
#endif


#ifndef HAVE_STRSPN
#define HAVE_STRSPN 0
#endif


#ifndef HAVE_STRVIS
#define HAVE_STRVIS 0
#endif


#ifndef HAVE_SYSCONF
#define HAVE_SYSCONF 0
#endif


#ifndef HAVE_VSYSLOG
#define HAVE_VSYSLOG 0
#endif


#ifndef HAVE_VWARNX
#define HAVE_VWARNX 0
#endif


#ifndef HAVE_CRYPT_H
#define HAVE_CRYPT_H 0
#endif


#ifndef HAVE_DIRENT_H
#define HAVE_DIRENT_H 0
#endif


#ifndef HAVE_FCNTL_H
#define HAVE_FCNTL_H 0
#endif


#ifndef HAVE_LIMITS_H
#define HAVE_LIMITS_H 0
#endif


#ifndef HAVE_MALLOC_H
#define HAVE_MALLOC_H 0
#endif


#ifndef HAVE_NDIR_H
#define HAVE_NDIR_H 0
#endif


#ifndef HAVE_NETINET_IP_H
#define HAVE_NETINET_IP_H 0
#endif


#ifndef HAVE_NETINET_IP_VAR_H
#define HAVE_NETINET_IP_VAR_H 0
#endif


#ifndef HAVE_PATHS_H
#define HAVE_PATHS_H 0
#endif


#ifndef HAVE_STRINGS_H
#define HAVE_STRINGS_H 0
#endif


#ifndef HAVE_SYS_DIR_H
#define HAVE_SYS_DIR_H 0
#endif


#ifndef HAVE_SYS_FILE_H
#define HAVE_SYS_FILE_H 0
#endif


#ifndef HAVE_SYS_IOCTL_H
#define HAVE_SYS_IOCTL_H 0
#endif


#ifndef HAVE_SYS_NDIR_H
#define HAVE_SYS_NDIR_H 0
#endif


#ifndef HAVE_SYS_TIME_H
#define HAVE_SYS_TIME_H 0
#endif


#ifndef HAVE_SYSLOG_H
#define HAVE_SYSLOG_H 0
#endif


#ifndef HAVE_TCPD_H
#define HAVE_TCPD_H 0
#endif


#ifndef HAVE_UNISTD_H
#define HAVE_UNISTD_H 0
#endif


#ifndef HAVE_LIBCRYPT
#define HAVE_LIBCRYPT 0
#endif


#ifndef HAVE_LIBDL
#define HAVE_LIBDL 0
#endif


#ifndef HAVE_LIBWRAP
#define HAVE_LIBWRAP 0
#endif


